import requests as r
import pandas as pd
from datetime import date, datetime, timedelta
from QualtricsAPI.Setup import Credentials
from QualtricsAPI.JSON import Parser

class Distributions(Credentials):
    '''This is a child class to the credentials class and gathers information about  Qualtric's Distributions.'''

    def __init__(self, token=None, directory_id=None, data_center=None):
        self.token = token
        self.data_center = data_center
        self.directory_id = directory_id
        return

    def set_send_date(self, weeks=0, days=0, hours=0, minutes=0, seconds=0):
        '''This method is a helper function to format the send date arguments for several methods in the Distribution Module.
        The send_date parameter must be in the format ""%Y-%m-%dT%H:%M:%SZ" in order for the API to properly parse the send date.
        Thus, this method defines the offset for the send_date, and formats it properly. An example would be if you wanted to send
        a reminder one week from now, simply pass "1" as an argument in to the "weeks" parameter. The default behaviour is for the
        send_date to be now, thus all params are set to zero offset.

        :param weeks: The week offset for the send_date. [Default = 0]
        :type weeks: int
        :param days: The day offset for the send_date. [Default = 0]
        :type days: int
        :param hours: The hour offset for the send_date. [Default = 0]
        :type hours: int
        :param minutes: The minute offset for the send_date. [Default = 0]
        :type minutes: int
        :param seconds: The second offset for the send_date. [Default = 0]
        :type seconds: int
        :return: The formatted DateTime. (str)
        '''
        send_date = datetime.now() + timedelta(weeks=weeks, days=days, hours=hours, minutes=minutes, seconds=seconds)
        return date.strftime(send_date, '%Y-%m-%dT%H:%M:%SZ')

    def create_distribution(self, subject, reply_email, from_email, from_name, mailing_list, library, survey, message, send_date, link_type='Individual'):
        '''This method gives users the ability to create a distribution for a given mailing list and survey. In order to use this method you
        must already have access to pre-defined Messages and their MessageID's existing within a User-Defined (starts with UR)
        or Global (starts with GR) Library. You can list the messages and their MessageID's(starts with MS)  that are available to your user
        account by using the "QualtricsAPI.Library.Messages.list_messages()" method. As a final note, this method gives users the ability to
        define the different types of messages.

        :param subject: The subject for the reminder email.
        :type subject: str
        :param reply_email: The reply email address.
        :type reply_email: str
        :param from_email: The email address that the distribution is sent from.
        :type from_email: str
        :param from_name: The name that shows up on the distribution.
        :type from_name: str
        :param library: The (Global or User) Library ID which the messages are located within.
        :type library: str
        :param message: The Message ID corresponding with the message that is to be sent.
        :type message: str
        :param mailing_list: The Mailing List ID corresponding with the Mailing List that the distribution is to be sent to.
        :type mailing_list: str
        :param survey: The Survey ID corresponding with the Survey that the distribution is to be sent to.
        :type survey: str
        :param send_date: The date that the distribution is supposed to be sent on. Pass gmtime() for immediate distribution or use the set_send_date() method to format properly.
        :type send_date: str
        :param link_type: This parameter refers to the type of link that is to be sent within the distribution.
        :type link_type: str
        :return: The Distribution ID. (str)
        '''

        assert len(mailing_list) == 18, 'Hey, the parameter for "mailing_list" that was passed is the wrong length. It should have 18 characters.'
        assert len(library) == 18, 'Hey, the parameter for "library" that was passed is the wrong length. It should have 18 characters.'
        assert len(survey) == 18, 'Hey, the parameter for "survey" that was passed is the wrong length. It should have 18 characters.'
        assert len(message) == 18, 'Hey, the parameter for "message" that was passed is the wrong length. It should have 18 characters.'
        assert mailing_list[:3] == 'CG_', 'Hey there! It looks like your Mailing List ID is incorrect. You can find the Mailing List ID on the Qualtrics site under your account settings. It will begin with "CG_". Please try again.'
        assert survey[:3] == 'SV_', 'Hey there! It looks like your SurveyID is incorrect. You can find the SurveyID on the Qualtrics site under your account settings. It will begin with "SV_". Please try again.'
        assert message[:3] == 'MS_', 'Hey there! It looks like your MessageID is incorrect. You can find the MessageID by using the list messages method available in the Messages module of this Package. It will begin with "MS_". Please try again.'
        assert library[:3] == 'UR_' or library[:3] == 'GR_', 'Hey there! It looks like your Library ID is incorrect. You can find the Library ID on the Qualtrics site under your account settings. It will begin with "UR_" or "GR_". Please try again.'

        headers, url = self.header_setup(content_type=True, xm=False, path='distributions')
        data = {
            'header': {
                'fromEmail': from_email,
                'fromName': from_name,
                'replyToEmail': reply_email,
                'subject': subject
            },
            'surveyLink': {
                'surveyId': survey,
                'type': link_type
            },
            'recipients': {
                'mailingListId': mailing_list
            },
            'sendDate': send_date,
            'message': {
                    'libraryId': library,
                    'messageId': message
            }
        }

        request = r.post(url, json=data, headers=headers)
        response = request.json()
        try:
            distribution_id = response['result']['id']
            return distribution_id
        except:
            print(f"\nServerError: QualtricsAPI Error Code: {response['meta']['error']['errorCode']}\nQualtricsAPI Error Message: {response['meta']['error']['errorMessage']}")
        return

    def create_reminder(self, subject, reply_email, from_email, from_name, library, message, distribution, send_date):
        '''This method gives users the ability to create a reminder for a given distribution. In order to create a reminders you must have
        already created a distribution for a given mailing list and survey. Once created, you will pass the Distribution ID corresponding to
        the correct distribution to the distribution parameter and the reminder will be set to that specific distribution. Unlike the
        QualtricsAPI.Survey.Distribution.create_distribution() method, this method does not require you to specify the mailing list or survey
        because it will use the parameters defined when the associated distribution was set up.

        :param subject: The subject for the reminder email.
        :type subject: str
        :param reply_email: The reply email address.
        :type reply_email: str
        :param from_email: The email address that the reminder is sent from.
        :type from_email: str
        :param from_name: The name that shows up on the reminder.
        :type from_name: str
        :param library: The (Global or User) Library ID which the messages are located within.
        :type library: str
        :param message: The Message ID corresponding with the message that is to be sent.
        :type message: str
        :param distribution: The Distribution ID corresponding with the distribution that the reminder is to be attached to.
        :type distribution: str
        :param send_date: The date that the reminder is supposed to be sent on. Pass gmtime() for immediate distribution or use the set_send_date() method to format properly.
        :type send_date: str
        :return: The "Reminder" Distribution ID. (str)
        '''

        assert len(distribution) == 19, 'Hey, the parameter for "distribution" that was passed is the wrong length. It should have 19 characters.'
        assert len(library) == 18, 'Hey, the parameter for "library" that was passed is the wrong length. It should have 18 characters.'
        assert len(message) == 18, 'Hey, the parameter for "message" that was passed is the wrong length. It should have 18 characters.'
        assert distribution[:4] == 'EMD_', 'Hey there! It looks like your distributionID is incorrect. You can find the distributionID by using the list_distributions method in this module. It will begin with "UMD_". Please try again.'
        assert message[:3] == 'MS_', 'Hey there! It looks like your MessageID is incorrect. You can find the MessageID by using the list messages method available in the Messages module of this Package. It will begin with "MS_". Please try again.'
        assert library[:3] == 'UR_' or library[:3] == 'GR_', 'Hey there! It looks like your Library ID is incorrect. You can find the Library ID on the Qualtrics site under your account settings. It will begin with "UR_" or "GR_". Please try again.'


        headers, base_url = self.header_setup(content_type=True, xm=False, path='distributions')
        url = f'{base_url}/{distribution}/reminders'
        data = {
            'header': {
                'fromEmail': from_email,
                'fromName': from_name,
                'replyToEmail': reply_email,
                'subject': subject
            },
            'sendDate': send_date,
            'message': {
                    'libraryId': library,
                    'messageId': message
            }
        }

        request = r.post(url, json=data, headers=headers)
        response = request.json()
        try:
            reminder_id = response['result']
            return reminder_id
        except:
            print(f"\nServerError: QualtricsAPI Error Code: {response['meta']['error']['errorCode']}\nQualtricsAPI Error Message: {response['meta']['error']['errorMessage']}")
        return

    def create_thank_you(self, subject, reply_email, from_email, from_name, library, message, distribution, send_date):
        '''This method gives users the ability to create a thank you for a given distribution. In order to create thank you distributions you
        must have already created a distribution for a given mailing list and survey. Once created, you will pass the Distribution ID
        corresponding to the correct distribution to the distribution parameter and the reminder will be set to that specific distribution.
        Unlike the  QualtricsAPI.Survey.Distribution.create_distribution() method, this method does not require you to specify the mailing
        list or survey because it will use the parameters defined when the associated distribution was set up.

        :param subject: The subject for the reminder email.
        :type subject: str
        :param reply_email: The reply email address.
        :type reply_email: str
        :param from_email: The email address that the reminder is sent from.
        :type from_email: str
        :param from_name: The name that shows up on the reminder.
        :type from_name: str
        :param library: The (Global or User) Library ID which the messages are located within.
        :type library: str
        :param message: The Message ID corresponding with the message that is to be sent.
        :type message: str
        :param distribution: The Distribution ID corresponding with the distribution that the reminder is to be attached to.
        :type distribution: str
        :param send_date: The date that the reminder is supposed to be sent on. Pass gmtime() for immediate distribution or use the set_send_date() method to format properly.
        :type send_date: str
        :return: The "Thank You" Distribution ID. (str)
        '''

        assert len(distribution) == 19, 'Hey, the parameter for "distribution" that was passed is the wrong length. It should have 19 characters.'
        assert len(library) == 18, 'Hey, the parameter for "library" that was passed is the wrong length. It should have 18 characters.'
        assert len(message) == 18, 'Hey, the parameter for "message" that was passed is the wrong length. It should have 18 characters.'
        assert distribution[:4] == 'EMD_', 'Hey there! It looks like your distributionID is incorrect. You can find the distributionID by using the list_distributions method in this module. It will begin with "UMD_". Please try again.'
        assert message[:3] == 'MS_', 'Hey there! It looks like your MessageID is incorrect. You can find the MessageID by using the list messages method available in the Messages module of this Package. It will begin with "MS_". Please try again.'
        assert library[:3] == 'UR_' or library[:3] == 'GR_', 'Hey there! It looks like your Library ID is incorrect. You can find the Library ID on the Qualtrics site under your account settings. It will begin with "UR_" or "GR_". Please try again.'


        headers, base_url = self.header_setup(content_type=True, xm=False, path='distributions')
        url = f'{base_url}/{distribution}/thankyous'
        data = {
            'header': {
                'fromEmail': from_email,
                'fromName': from_name,
                'replyToEmail': reply_email,
                'subject': subject
            },
            'sendDate': send_date,
            'message': {
                    'libraryId': library,
                    'messageId': message
            }
        }
        request = r.post(url, json=data, headers=headers)
        response = request.json()
        try:
            thanks_id = response['result']
            return thanks_id
        except:
            print(f"\nServerError: QualtricsAPI Error Code: {response['meta']['error']['errorCode']}\nQualtricsAPI Error Message: {response['meta']['error']['errorMessage']}")
        return

    def list_distributions(self, survey):
        ''' This method will list all of the distributions corresponding with a given survey. Given that distributions are
        specific to individual surveys, we must pass the SurveyID as an arguement into the survey parameter for this method
        to work appropriately. This method will return a Pandas DataFrame filled with a list of the distributions associated
        with the specific SurveyID passed to the survey parameter.

        :param survey: The Survey ID corresponding with the Survey that the distribution is to be sent to.
        :type survey: str
        :return: A Pandas DataFrame
        '''

        assert survey[:3] == 'SV_', 'Hey there! It looks like your SurveyID is incorrect. You can find the SurveyID on the Qualtrics site under your account settings. It will begin with "SV_". Please try again.'
        assert len(survey) == 18, 'Hey, the parameter for "survey" that was passed is the wrong length. It should have 18 characters.'

        headers, base_url = self.header_setup(xm=False, path='distributions')
        url = f'{base_url}?surveyId={survey}'
        columns = ['id', 'parentDistributionId', 'ownerId', 'organizationId', 'requestStatus', 'requestType',
                    'sendDate', 'createdDate', 'modifiedDate', 'headers', 'fromEmail', 'replyToEmail', 'fromName',
                     'subject', 'recipients', 'mailingListId', 'contactId', 'sampleId', 'message', 'messageId',
                     'messageText', 'surveyLink', 'surveyId', 'expirationDate', 'linkType', 'stats', 'sent', 'failed',
                     'started', 'bounced', 'opened', 'skipped', 'finished', 'complaints', 'blocked', 'mailing_list_library_id', 'message_library_id']
        master = pd.DataFrame(columns=columns)
        def extract_distributions(url=url, master=master):
            request = r.get(url, headers=headers)
            response = request.json()
            if response['meta']['httpStatus'] == '200 - OK':
                keys = columns[:-2]
                dists = Parser().json_parser(response=response, keys=keys, arr=False)
                dist_df = pd.DataFrame(dists).transpose()
                dist_df.columns = keys
                library_ids = Parser().json_parser(response=response, keys=['libraryId'], arr=False)
                dist_df['mailing_list_library_id'] = library_ids[0][:len(dist_df)]
                dist_df['message_library_id'] = library_ids[0][len(dist_df):]
                master = pd.concat([master, dist_df], sort=False).reset_index(drop=True)
                next_page = response['result']['nextPage']
                return master, next_page
            else:
                print(response['meta'])
                master, next_page = extract_distributions(url=url, master=master)

        master, next_page = extract_distributions()
        if next_page == None:
            return master
        else:
            while next_page != None:
                master, next_page = extract_distributions(url=next_page, master=master)
            return master
                

    def get_distribution(self, survey, distribution):
        ''' This method gives users the ability to get a specific distribution corresponding with a given survey. Given that
        distributions are specific to individual surveys, we must pass the SurveyID as an arguement into the survey parameter for
        this method to work appropriately. This method will return a Pandas DataFrame consisting of multiple variables associated
        with the specified distirbution.

        :param survey: The Survey ID corresponding with the Survey that the distribution is to be sent to.
        :type survey: str
        :param distribution: A specific Distribution ID associated with the given survey.
        :type distribution: str
        :return: A Pandas DataFrame
        '''

        assert survey[:3] == 'SV_', 'Hey there! It looks like your SurveyID is incorrect. You can find the SurveyID on the Qualtrics site under your account settings. It will begin with "SV_". Please try again.'
        assert len(survey) == 18, 'Hey, the parameter for "survey" that was passed is the wrong length. It should have 18 characters.'
        assert len(distribution) == 19, 'Hey, the parameter for "distribution" that was passed is the wrong length. It should have 19 characters.'
        assert distribution[:4] == 'EMD_', 'Hey there! It looks like your distributionID is incorrect. You can find the distributionID by using the list_distributions method in this module. It will begin with "UMD_". Please try again.'

        headers, base_url = self.header_setup(xm=False, path='distributions')
        url = f'{base_url}/{distribution}?surveyId={survey}'
        request = r.get(url, headers=headers)
        try:
            response = request.json()
            keys = ['id', 'parentDistributionId', 'ownerId', 'organizationId', 'requestStatus', 'requestType',
                    'sendDate', 'createdDate', 'modifiedDate', 'headers', 'fromEmail', 'replyToEmail', 'fromName',
                     'subject', 'recipients', 'mailingListId', 'contactId', 'sampleId', 'message', 'messageId',
                     'messageText', 'surveyLink', 'surveyId', 'expirationDate', 'linkType', 'stats', 'sent', 'failed',
                     'started', 'bounced', 'opened', 'skipped', 'finished', 'complaints', 'blocked']
            dists = Parser().json_parser(response=response, keys=keys, arr=False)
            dist_df = pd.DataFrame(dists)
            dist_df.index = keys
            library_ids = Parser().json_parser(response=response, keys=['libraryId'], arr=False)
            lib = pd.DataFrame(library_ids[0], index=['mailing_list_library_id', 'message_library_id'])
            dist_df = dist_df.append(lib)
            return dist_df
        except:
            print(f"\nServerError: QualtricsAPI Error Code: {response['meta']['error']['errorCode']}\nQualtricsAPI Error Message: {response['meta']['error']['errorMessage']}")


    def create_sms_distribution(self, dist_name, mailing_list, library, survey, message, send_date, parentDistributionId=None, method='Invite'):
        '''This method gives users the ability to create a SMS distribution for a given mailing list and survey. In order to use this method you
        must already have access to pre-defined Messages and their MessageID's existing within a User-Defined (starts with UR)
        or Global (starts with GR) Library. You can list the messages and their MessageID's(starts with MS)  that are available to your user
        account by using the "QualtricsAPI.Library.Messages.list_messages()" method. 

        :param dist_name: The name that shows up for the distribution.
        :type dist_name: str
        :param library: The (Global or User) Library ID which the messages are located within.
        :type library: str
        :param message: The Message ID corresponding with the message that is to be sent.
        :type message: str
        :param mailing_list: The Mailing List ID corresponding with the Mailing List that the distribution is to be sent to.
        :type mailing_list: str
        :param survey: The Survey ID corresponding with the Survey that the distribution is to be sent to.
        :type survey: str
        :param send_date: The date that the distribution is supposed to be sent on. Pass gmtime() for immediate distribution or use the set_send_date() method to format properly.
        :type send_date: str
        :param method: This parameter refers to the type of distribution that is to be sent out to the mailing list.
        :type method: str
        :return: The Distribution ID. (str)
        '''

        assert len(mailing_list) == 18, 'Hey, the parameter for "mailing_list" that was passed is the wrong length. It should have 18 characters.'
        assert len(library) == 18, 'Hey, the parameter for "library" that was passed is the wrong length. It should have 18 characters.'
        assert len(survey) == 18, 'Hey, the parameter for "survey" that was passed is the wrong length. It should have 18 characters.'
        assert len(message) == 18, 'Hey, the parameter for "message" that was passed is the wrong length. It should have 18 characters.'
        assert mailing_list[:3] == 'CG_', 'Hey there! It looks like your Mailing List ID is incorrect. You can find the Mailing List ID on the Qualtrics site under your account settings. It will begin with "CG_". Please try again.'
        assert survey[:3] == 'SV_', 'Hey there! It looks like your SurveyID is incorrect. You can find the SurveyID on the Qualtrics site under your account settings. It will begin with "SV_". Please try again.'
        assert message[:3] == 'MS_', 'Hey there! It looks like your MessageID is incorrect. You can find the MessageID by using the list messages method available in the Messages module of this Package. It will begin with "MS_". Please try again.'
        assert library[:3] == 'UR_' or library[:3] == 'GR_', 'Hey there! It looks like your Library ID is incorrect. You can find the Library ID on the Qualtrics site under your account settings. It will begin with "UR_" or "GR_". Please try again.'

        headers, url = self.header_setup(content_type=True, xm=False, path='distributions/sms')
        data = {
            'sendDate': send_date,
            'surveyId': survey,
            'method': method,
            'recipients': {
                'mailingListId': mailing_list
            },
            'name': dist_name,
            'message': {
                    'libraryId': library,
                    'messageId': message
            }
        }

        if parentDistributionId != None: 
           data['parentDistributionId'] = parentDistributionId

        request = r.post(url, json=data, headers=headers)
        response = request.json()
        try:
            distribution_id = response['result']['id']
            return distribution_id
        except:
            print(f"\nServerError: QualtricsAPI Error Code: {response['meta']['error']['errorCode']}\nQualtricsAPI Error Message: {response['meta']['error']['errorMessage']}")
        return


# __init__.py
from QualtricsAPI.Setup import Credentials
from QualtricsAPI.Survey import Responses
from QualtricsAPI.JSON import Parser
from QualtricsAPI.XM import MailingList
from QualtricsAPI.XM import XMDirectory
from QualtricsAPI.Library import Messages

__all__ = ["Setup", "JSON", "Contacts", "Survey", "Library"]



import numpy as np

class Parser(object):

    def __init__(self):
        return

    def extract_values(self, obj=None, key=None):
        ''' This outer method  a specific value from a nested dictionary.

        :param obj: a dictionary object.
        :param key: The Specific Key that the value is associated with.
        :return: the value associated with the given key
        '''
        values = []
        def extract(obj, values, key):
            '''This inner method recursively searches for the given key in a nested dictionary. (Not a User-Facing Method)

            :param obj: a dictionary object.
            :param values: a list that will house the values of the given key.
            :param key: The Specific Key that the value is associated with.
            :return: the value associated with the given key
            '''
            if isinstance(obj, dict):
                for k, v in obj.items():
                    if isinstance(v, (dict, list)):
                        extract(v, values, key)
                    elif k == key:
                        values.append(v)
            elif isinstance(obj, list):
                for j in obj:
                    extract(j, values, key)
            return values
        results = extract(obj, values, key)
        return results

    def extract_keys(self, obj):
        '''This outer method extracts all of the keys from a nested dictionary.

        :param obj: a dictionary object.
        :return: a list of keys within a given dictonary.
        '''
        keys = []

        def extract(obj, keys):
            '''This inner method recursively locates each of the keys within a nested dictionary.

            :param obj: a dictionary object.
            :param keys: a list of previously identified keys
            :return: a list of keys within a given recursion of the inner method.
            '''
            if isinstance(obj, dict):
                for k,v in obj.items():
                    if isinstance(v, (dict, list)):
                        keys.append(k)
                        extract(v, keys)
                    else:
                        keys.append(k)
            elif isinstance(obj, list):
                for element in obj:
                    extract(element, keys)
            return keys
        obj_keys =  extract(obj, keys)
        return obj_keys

    def json_parser(self, response=None, keys=[], arr=True):
        '''This method itterates over the all of the keys within a keys list.'''

        #Improvement: Include the Extract keys method in this method.
        elements  = [self.extract_values(response, item) for item in keys]
        if arr == True:
            return np.array(elements).T
        else:
            return elements



# Run in the Root Directory to run tests
## python3 -m unittest QualtricsAPI/tests/test.py

import unittest
import pandas as pd
from QualtricsAPI.Setup import Credentials
from QualtricsAPI.Survey import Responses
from QualtricsAPI.JSON import Parser
from QualtricsAPI.XM import MailingList
from QualtricsAPI.XM import XMDirectory
from QualtricsAPI.Library import Messages
from QualtricsAPI.Survey import Distributions
from datetime import date, datetime, timedelta
from time import gmtime

# Setup Tests Class
class setup_tests(object):

    def __init__(self):
        return

    def setup_test_token(self, short=False):
        '''Setup for Test Case 1: qualtrics_api_credentials token parameter lengths.(40)'''

        token = 'ThisIsaFakeAPITokenAndIsTooShortToWork!' if short else 'ThisIsaFakeAPITokenAndIsTooShortToWork!!!'
        return token

    def setup_test_directory_id(self, short=False, false_id=False):
        '''Setup for Test Case 2: qualtrics_api_credentials directory id parameter lengths (20), and the incorrect id (POOL_). '''

        directory_id = 'POOL_ThisIsaFakeID!' if short else 'POOL_ThisIsaFakeDirectoryID!'
        bad_id = 'ThisIsaFakeIDwo/POOL' if false_id else 'POOL_ThisIsaFakeID!'
        return directory_id, bad_id

    def setup_test_mailing_list_id(self, short=False, false_id=False):
        '''Setup for Test Case 3: Mailing List Sub-Module method's exception handling of the mailing list's length (18), and
        the incorrect id (CG_).'''

        mailing_list_id = 'CG_ThisIsaFakeID!' if short else 'CG_ThisIsaFakeMailingID!'
        bad_id = 'ThisIsaFakeIDwo/CG' if false_id else None
        return mailing_list_id, bad_id

    def setup_test_contact_id(self, short=False, false_id=False):
        '''Setup for Test Case 4: XMDirectory Sub-Module method's exception handling of the contact_id's length (18), and
        the incorrect id (CG_).'''

        contact_id = 'CID_ThisIsaFakeID!' if short else 'CID_ThisIsaFakeMailingID!'
        bad_id = 'ThisIsaFakeIDwo/CID' if false_id else None
        return contact_id, bad_id

    def setup_test_survey_id(self, short=False, false_id=False):
        '''Setup for Test Case 5: Responses Sub-Module method's exception handling of the survey_id's length (18), and
        the incorrect id (SV_).'''

        survey_id = 'SV_ThisIsaFakeID!' if short else 'SV_ThisIsaFakeMailingID!'
        bad_id = 'ThisIsaFakeIDwo/SV' if false_id else None
        return survey_id, bad_id

    def setup_test_library_id_ur(self, short=False, false_id=False):
        '''Setup for Test Case 6: Responses Sub-Module method's exception handling of the library_id's length (18), and
        the incorrect id (UR_).'''

        lib_id = 'UR_ThisIsaFakeID!' if short else 'UR_ThisIsaFakeMailingID!'
        bad_id = 'ThisIsaFakeIDwo/UR' if false_id else None
        return lib_id, bad_id

    def setup_test_library_id_gr(self, short=False, false_id=False):
        '''Setup for Test Case 7: Responses Sub-Module method's exception handling of the library_id's length (18), and
        the incorrect id (GR_).'''

        lib_id = 'GR_ThisIsaFakeID!' if short else 'GR_ThisIsaFakeMailingID!'
        bad_id = 'ThisIsaFakeIDwo/GR' if false_id else None
        return lib_id, bad_id

    def setup_test_dist_id(self, short=False, false_id=False):
        '''Setup for Test Case 8: Responses Sub-Module method's exception handling of the distribution_id's length (19), and
        the incorrect id (UMD_).'''

        dist_id = 'EMD_ThisIsaFakeID!' if short else 'EMD_ThisIsaFakeMailingID!'
        bad_id = 'ThisIsaFakeIDwo/EMD' if false_id else None
        return dist_id, bad_id

    def setup_test_message_id(self, short=False, false_id=False):
        '''Setup for Test Case 9: Responses Sub-Module method's exception handling of the message_id's length (18), and
        the incorrect id (MS_).'''

        msg_id = 'MS_ThisIsaFakeID!' if short else 'MS_ThisIsaFakeMailingID!'
        bad_id = 'ThisIsaFakeIDwo/MS' if false_id else None
        return msg_id, bad_id

#UnitTest Class
class TestQualtricsAPI(unittest.TestCase):

    correct_token = 'ThisIsaFakeAPITokenAndIsTooShortToWork!!'

    def test_credentials_long_token(self):
        '''This method tests that an assertion is raised in the Credentials Module when the user enters an api token that is too long.'''
        token = setup_tests().setup_test_token(short=False)
        directory_id, bad_id = setup_tests().setup_test_directory_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Credentials().qualtrics_api_credentials(token=token, data_center='FAKE', directory_id=directory_id)

    #Test Assertion Error is handled: Short Token
    def test_credentials_short_token(self):
        '''This method tests that an assertion is raised in the Credentials Module when the user enters an api token that is too short.'''
        token = setup_tests().setup_test_token(short=True)
        directory_id, bad_id = setup_tests().setup_test_directory_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Credentials().qualtrics_api_credentials(token=token, data_center='FAKE', directory_id=directory_id)

    ## API Credentials: Directory ID
    #Test Assertion Error is handled: Short Directory id
    def test_credentials_short_directory_id(self):
        '''This method tests that an assertion is raised in the Credentials Module when the user enters a directory id that is too short.'''
        token = setup_tests().setup_test_token(short=True)
        directory_id, bad_id = setup_tests().setup_test_directory_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Credentials().qualtrics_api_credentials(token=TestQualtricsAPI.correct_token, data_center='FAKE', directory_id=directory_id)

    #Test Assertion Error is handled: Long Directory id
    def test_credentials_long_directory_id(self):
        '''This method tests that an assertion is raised in the Credentials Module when the user enters a directory id that is too long.'''
        token = setup_tests().setup_test_token(short=True)
        directory_id, bad_id = setup_tests().setup_test_directory_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Credentials().qualtrics_api_credentials(token=TestQualtricsAPI.correct_token, data_center='FAKE', directory_id=directory_id)

    #Test Assertion Error is handled: Incorrect Dictonary ID
    def test_credentials_bad_directory_id(self):
        '''This method tests that an assertion is raised in the Credentials Module when the user enters a directory id that is incorrect.'''
        token = setup_tests().setup_test_token(short=True)
        directory_id, bad_id = setup_tests().setup_test_directory_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Credentials().qualtrics_api_credentials(token=TestQualtricsAPI.correct_token, data_center='FAKE', directory_id=bad_id)


    ## MailingList: Mailing List IDs (rename_list)
    #Test Assertion Error is handled: Long Mailing List ID
    def test_ml_short_ml_id_rename(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is too long.'''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            MailingList().rename_list(mailing_list=mailing_list_id, name='Fake')

    #Test Assertion Error is handled: Short Mailing List ID
    def test_ml_long_ml_id_rename(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is too short.'''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            MailingList().rename_list(mailing_list=mailing_list_id, name='Fake')

    #Test Assertion Error is handled: Incorrect Mailing List ID
    def test_ml_bad_ml_id_rename(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is incorrect. '''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            MailingList().rename_list(mailing_list=bad_id, name='Fake')

    #Test Assertion Error is handled: (None) Mailing List ID
    def test_ml_none_ml_id_rename(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is incorrect. '''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            MailingList().rename_list(mailing_list=bad_id, name='Fake')

    #Test Assertion Error is handled: bool Mailing List ID
    def test_ml_bool_ml_id_rename(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is incorrect. '''
        NonString = bool()
        with self.assertRaises(AssertionError):
            MailingList().rename_list(mailing_list=NonString, name='Fake')

    #Test Assertion Error is handled: bool Mailing List ID
    def test_ml_list_ml_id_rename(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is incorrect. '''
        NonString = list()
        with self.assertRaises(AssertionError):
            MailingList().rename_list(mailing_list=NonString, name='Fake')

    #Test Assertion Error is handled: float Mailing List ID
    def test_ml_float_ml_id_rename(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is incorrect. '''
        NonString = float()
        with self.assertRaises(AssertionError):
            MailingList().rename_list(mailing_list=NonString, name='Fake')

    #Test Assertion Error is handled: int Mailing List ID
    def test_ml_int_ml_id_rename(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is incorrect. '''
        NonString = int()
        with self.assertRaises(AssertionError):
            MailingList().rename_list(mailing_list=NonString, name='Fake')

    #Test Assertion Error is handled: dict Mailing List ID
    def test_ml_dict_ml_id_rename(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is incorrect. '''
        NonString = dict()
        with self.assertRaises(AssertionError):
            MailingList().rename_list(mailing_list=NonString, name='Fake')

    #Test Assertion Error is handled: tuple Mailing List ID
    def test_ml_tuple_ml_id_rename(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is incorrect. '''
        NonString = tuple()
        with self.assertRaises(AssertionError):
            MailingList().rename_list(mailing_list=NonString, name='Fake')

    ## MailingList: Mailing List IDs (delete_list)
    #Test Assertion Error is handled: Long Mailing List ID
    def test_ml_short_ml_id_delete(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is too long.'''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            MailingList().delete_list(mailing_list=mailing_list_id)

    #Test Assertion Error is handled: Short Mailing List ID
    def test_ml_long_ml_id_delete(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is too short.'''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            MailingList().delete_list(mailing_list=mailing_list_id)

    #Test Assertion Error is handled: None Mailing List ID
    def test_ml_none_ml_id_delete(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is too short.'''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=True, false_id=True)
        with self.assertRaises(AssertionError):
            MailingList().delete_list(mailing_list=mailing_list_id)

    #Test Assertion Error is handled: bool Mailing List ID
    def test_ml_bool_ml_id_delete(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is too short.'''
        NonString = bool()
        with self.assertRaises(AssertionError):
            MailingList().delete_list(mailing_list=NonString)

    #Test Assertion Error is handled: Incorrect Mailing List ID
    def test_ml_bad_ml_id_delete(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id is that is incorrect.'''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            MailingList().delete_list(mailing_list=bad_id)

    ## MailingList: Mailing List IDs (list_contacts)
    #Test Assertion Error is handled: Long Mailing List ID
    def test_ml_short_ml_id_list_contacts(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is too long.'''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            MailingList().list_contacts(mailing_list=mailing_list_id)

    #Test Assertion Error is handled: Short Mailing List ID
    def test_ml_long_ml_id_list_contacts(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is too short. '''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            MailingList().list_contacts(mailing_list=mailing_list_id)

    #Test Assertion Error is handled: Incorrect Mailing List ID
    def test_ml_bad_ml_id_list_contacts(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is incorrect. '''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            MailingList().list_contacts(mailing_list=bad_id)

    ## MailingList: Mailing List IDs (create_contact_in_list)
    #Test Assertion Error is handled: Long Mailing List ID
    def test_ml_short_ml_id_create(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is too long.'''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            MailingList().create_contact_in_list(mailing_list=mailing_list_id)

    #Test Assertion Error is handled: Short Mailing List ID
    def test_ml_long_ml_id_create(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id that is too short.'''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            MailingList().create_contact_in_list(mailing_list=mailing_list_id)

    #Test Assertion Error is handled: Incorrect Mailing List ID
    def test_ml_bad_ml_id_create(self):
        '''This method tests that an assertion is raised in the MailingList Module when the user enters a mailing_list_id is that is incorrect.'''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            MailingList().create_contact_in_list(mailing_list=bad_id)

    ## XMDirectory: Contact IDs (delete_contact)
    #Test Assertion Error is handled: Long Contact id
    def test_xm_short_contact_id_delete(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is too long. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().delete_contact(contact_id=contact_id)

    #Test Assertion Error is handled: Short Contact id
    def test_xm_long_contact_id_delete(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is too short. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().delete_contact(contact_id=contact_id)

    #Test Assertion Error is handled: Incorrect Contact id
    def test_xm_bad_contact_id_delete(self):
        '''This method tests that an assertion is raised in the  XMDirectory Module when the user enters a contact_id that is incorrect. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            XMDirectory().delete_contact(contact_id=bad_id)

    #Test Assertion Error is handled: (None) Contact id
    def test_xm_none_contact_id_delete(self):
        '''This method tests that an assertion is raised in the  XMDirectory Module when the user enters a contact_id that is incorrect. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().delete_contact(contact_id=bad_id)

    #Test Assertion Error is handled: bool Contact id
    def test_xm_bool_contact_id_delete(self):
        '''This method tests that an assertion is raised in the  XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = bool()
        with self.assertRaises(AssertionError):
            XMDirectory().delete_contact(contact_id=NonString)

    #Test Assertion Error is handled: list Contact id
    def test_xm_list_contact_id_delete(self):
        '''This method tests that an assertion is raised in the  XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = list()
        with self.assertRaises(AssertionError):
            XMDirectory().delete_contact(contact_id=NonString)

    #Test Assertion Error is handled: int Contact id
    def test_xm_int_contact_id_delete(self):
        '''This method tests that an assertion is raised in the  XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = int()
        with self.assertRaises(AssertionError):
            XMDirectory().delete_contact(contact_id=NonString)

    #Test Assertion Error is handled: float Contact id
    def test_xm_float_contact_id_delete(self):
        '''This method tests that an assertion is raised in the  XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = float()
        with self.assertRaises(AssertionError):
            XMDirectory().delete_contact(contact_id=NonString)

    #Test Assertion Error is handled: dict Contact id
    def test_xm_dict_contact_id_delete(self):
        '''This method tests that an assertion is raised in the  XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = dict()
        with self.assertRaises(AssertionError):
            XMDirectory().delete_contact(contact_id=NonString)

    #Test Assertion Error is handled: tuple Contact id
    def test_xm_tuple_contact_id_delete(self):
        '''This method tests that an assertion is raised in the  XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = tuple()
        with self.assertRaises(AssertionError):
            XMDirectory().delete_contact(contact_id=NonString)

    ## XMDirectory: Contact IDs (update_contact)
    #Test Assertion Error is handled: Long Contact id
    def test_xm_short_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is too long. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=contact_id)

    #Test Assertion Error is handled: Short Contact id
    def test_xm_long_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is too short. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=contact_id)

    #Test Assertion Error is handled: Incorrect Contact id
    def test_xm_bad_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=bad_id)

    #Test Assertion Error is handled: (None) Contact id
    def test_xm_bad_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=bad_id)

    #Test Assertion Error is handled: (None) Contact id
    def test_xm_bad_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=bad_id)

    #Test Assertion Error is handled: (None) Contact id
    def test_xm_none_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=bad_id)

    #Test Assertion Error is handled: bool Contact id
    def test_xm_bool_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = bool()
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=NonString)

    #Test Assertion Error is handled: list Contact id
    def test_xm_list_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = list()
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=NonString)

    #Test Assertion Error is handled: int Contact id
    def test_xm_int_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = int()
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=NonString)

    #Test Assertion Error is handled: float Contact id
    def test_xm_float_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = float()
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=NonString)

    #Test Assertion Error is handled: dict Contact id
    def test_xm_dict_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = dict()
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=NonString)

    #Test Assertion Error is handled: tuple Contact id
    def test_xm_tuple_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = tuple()
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=NonString)

    #Test Assertion Error is handled: (None) Contact id
    def test_xm_bad_contact_id_update(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().update_contact(contact_id=bad_id)

    ## XMDirectory: Contact IDs (get_contact)
    #Test Assertion Error is handled: Long Contact id
    def test_xm_short_contact_id_get(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is too long. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact(contact_id=contact_id)

    #Test Assertion Error is handled: Short Contact id
    def test_xm_long_contact_id_get(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is too short. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact(contact_id=contact_id)

    #Test Assertion Error is handled: Incorrect Contact id
    def test_xm_bad_contact_id_get(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact(contact_id=bad_id)

    #Test Assertion Error is handled: (None) Contact id
    def test_xm_none_contact_id_get(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact(contact_id=bad_id)

    #Test Assertion Error is handled: bool Contact id
    def test_xm_bool_contact_id_get(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        bad_id = bool()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact(contact_id=bad_id)

    #Test Assertion Error is handled: list Contact id
    def test_xm_list_contact_id_get(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        bad_id = list()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact(contact_id=bad_id)

    #Test Assertion Error is handled: float Contact id
    def test_xm_float_contact_id_get(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        bad_id = float()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact(contact_id=bad_id)

    #Test Assertion Error is handled: dict Contact id
    def test_xm_dict_contact_id_get(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        bad_id = dict()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact(contact_id=bad_id)

    #Test Assertion Error is handled: tuple Contact id
    def test_xm_tuple_contact_id_get(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        bad_id = tuple()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact(contact_id=bad_id)

    #Test Assertion Error is handled: int Contact id
    def test_xm_int_contact_id_get(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        bad_id = int()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact(contact_id=bad_id)

    ## XMDirectory: Contact IDs (get_contact_additional_info)
    #Test Assertion Error is handled: Long Contact id
    def test_xm_short_contact_id_get_add(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is too long. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact_additional_info(contact_id=contact_id)

    #Test Assertion Error is handled: Short Contact id
    def test_xm_long_contact_id_get_add(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is too short. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact_additional_info(contact_id=contact_id)

    #Test Assertion Error is handled: Incorrect Contact id
    def test_xm_bad_contact_id_get_add(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact_additional_info(contact_id=bad_id)

    #Test Assertion Error is handled: (None) Contact id
    def test_xm_none_contact_id_get_add(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        contact_id, bad_id = setup_tests().setup_test_contact_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact_additional_info(contact_id=bad_id)

    #Test Assertion Error is handled: (None) Content
    def test_xm_none_content_get_add(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact_additional_info(contact_id='ThisisaFakeContact', content=None)

    #Test Assertion Error is handled: bool Contact id
    def test_xm_bool_contact_id_get_add(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = bool()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact_additional_info(contact_id=NonString)

    #Test Assertion Error is handled: int Contact id
    def test_xm_int_contact_id_get_add(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = int()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact_additional_info(contact_id=NonString)

    #Test Assertion Error is handled: list Contact id
    def test_xm_list_contact_id_get_add(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = list()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact_additional_info(contact_id=NonString)

    #Test Assertion Error is handled: float Contact id
    def test_xm_float_contact_id_get_add(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = float()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact_additional_info(contact_id=NonString)

    #Test Assertion Error is handled: dict Contact id
    def test_xm_dict_contact_id_get_add(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = dict()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact_additional_info(contact_id=NonString)

    #Test Assertion Error is handled: tuple Contact id
    def test_xm_tuple_contact_id_get_add(self):
        '''This method tests that an assertion is raised in the XMDirectory Module when the user enters a contact_id that is incorrect. '''
        NonString = tuple()
        with self.assertRaises(AssertionError):
            XMDirectory().get_contact_additional_info(contact_id=NonString)

    ## Responses: Survey IDs ##
    #Test Assertion Error is handled: Long Survey id
    def test_responses_short_survey_id(self):
        '''This method tests that an assertion is raised in the Responses Module when the user enters a survey_id that is too long. '''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Responses().setup_request(file_format='csv', survey=survey_id)

    #Test Assertion Error is handled: Short Survey id
    def test_responses_long_survey_id(self):
        '''This method tests that an assertion is raised in the Responses Module when the user enters a survey_id that is too short. '''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Responses().setup_request(file_format='csv', survey=survey_id)

    #Test Assertion Error is handled: Incorrect Survey id
    def test_responses_bad_survey_id(self):
        '''This method tests that an assertion is raised in the Responses Module when the user enters a survey_id that is incorrect. '''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Responses().setup_request(file_format='csv', survey=bad_id)

    #Test Assertion Error is handled: (None) Survey id
    def test_responses_none_survey_id(self):
        '''This method tests that an assertion is raised in the Responses Module when the user enters a survey_id that is incorrect. '''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Responses().setup_request(file_format='csv', survey=bad_id)

    #Test Assertion Error is handled: bool Survey id
    def test_responses_bool_survey_id(self):
        '''This method tests that an assertion is raised in the Responses Module when the user enters a survey_id that is incorrect. '''
        NonString = bool()
        with self.assertRaises(AssertionError):
            Responses().setup_request(file_format='csv', survey=NonString)

    #Test Assertion Error is handled: list Survey id
    def test_responses_list_survey_id(self):
        '''This method tests that an assertion is raised in the Responses Module when the user enters a survey_id that is incorrect. '''
        NonString = list()
        with self.assertRaises(AssertionError):
            Responses().setup_request(file_format='csv', survey=NonString)

    #Test Assertion Error is handled: int Survey id
    def test_responses_int_survey_id(self):
        '''This method tests that an assertion is raised in the Responses Module when the user enters a survey_id that is incorrect. '''
        NonString = int()
        with self.assertRaises(AssertionError):
            Responses().setup_request(file_format='csv', survey=NonString)

    #Test Assertion Error is handled: float Survey id
    def test_responses_float_survey_id(self):
        '''This method tests that an assertion is raised in the Responses Module when the user enters a survey_id that is incorrect. '''
        NonString = float()
        with self.assertRaises(AssertionError):
            Responses().setup_request(file_format='csv', survey=NonString)

    #Test Assertion Error is handled: dict Survey id
    def test_responses_dict_survey_id(self):
        '''This method tests that an assertion is raised in the Responses Module when the user enters a survey_id that is incorrect. '''
        NonString = dict()
        with self.assertRaises(AssertionError):
            Responses().setup_request(file_format='csv', survey=NonString)

    #Test Assertion Error is handled: tuple Survey id
    def test_responses_tuple_survey_id(self):
        '''This method tests that an assertion is raised in the Responses Module when the user enters a survey_id that is incorrect. '''
        NonString = tuple()
        with self.assertRaises(AssertionError):
            Responses().setup_request(file_format='csv', survey=NonString)

    ## Messages: Library IDs ##
    #Test Assertion Error is handled: Long Library id (list_messages: UR)
    def test_responses_short_library_id_list_msgs_ur(self):
        '''This method tests that an assertion is raised in the Messages Module's list_messages method when the user enters a
        library_id (UR: User-Defined Resource) that is too long. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_ur(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Messages().list_messages(library=lib_id)

    #Test Assertion Error is handled: Short Library id (list_messages: UR)
    def test_responses_long_library_id_lst_msgs_ur(self):
        '''This method tests that an assertion is raised in the Messages Module's list_messages method when the user enters a
         library_id (UR: User-Defined Resource) that is too short. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_ur(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Messages().list_messages(library=lib_id)

    #Test Assertion Error is handled: Incorrect Library id (list_messages: UR)
    def test_responses_bad_libary_id_lst_msgs_ur(self):
        '''This method tests that an assertion is raised in the Messages Module's list_messages method when the user enters
         a library_id (UR: User-Defined Resource) that is incorrect. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_ur(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Messages().list_messages(library=bad_id)

    #Test Assertion Error is handled: Long Library id (list_messages: GR)
    def test_responses_short_library_id_list_msgs_gr(self):
        '''This method tests that an assertion is raised in the Messages Module's list_messages method when the user enters
         a library_id (GR: Global Resource) that is too long. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_gr(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Messages().list_messages(library=lib_id)

    #Test Assertion Error is handled: Short Library id (list_messages: GR)
    def test_responses_long_library_id_lst_msgs_gr(self):
        '''This method tests that an assertion is raised in the Messages Module's list_messages method when the user enters
         a library_id (GR: Global Resource) that is too short. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_gr(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Messages().list_messages(library=lib_id)

    #Test Assertion Error is handled: Incorrect Library id (list_messages: GR)
    def test_responses_bad_libary_id_lst_msgs_gr(self):
        '''This method tests that an assertion is raised in the Messages Module's list_messages method when the user enters
         a library_id (GR: Global Resource) that is incorrect. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_gr(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Messages().list_messages(library=bad_id)

    ## Messages: Library IDs ##
    #Test Assertion Error is handled: Long Library id (get_message: UR)
    def test_responses_short_library_id_get_msg_ur(self):
        '''This method tests that an assertion is raised in the Messages Module's get_message method when the user enters a
        library_id (UR: User-Defined Resource) that is too long. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_ur(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Messages().get_message(library=lib_id, message="MS_ThisIsaFakeMsg!")

    #Test Assertion Error is handled: Short Library id (get_message: UR)
    def test_responses_long_library_id_get_msg_ur(self):
        '''This method tests that an assertion is raised in the Messages Module's get_message method when the user enters a
         library_id (UR: User-Defined Resource) that is too short. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_ur(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Messages().get_message(library=lib_id, message="MS_ThisIsaFakeMsg!")

    #Test Assertion Error is handled: Incorrect Library id (get_message: UR)
    def test_responses_bad_libary_id_get_msg_ur(self):
        '''This method tests that an assertion is raised in the Messages Module's get_message method when the user enters
         a library_id (UR: User-Defined Resource) that is incorrect. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_ur(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Messages().get_message(library=bad_id, message="MS_ThisIsaFakeMsg!")

    #Test Assertion Error is handled: Long Library id (get_message: GR)
    def test_responses_short_library_id_get_msg_gr(self):
        '''This method tests that an assertion is raised in the Messages Module's get_message method when the user enters
         a library_id (GR: Global Resource) that is too long. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_gr(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Messages().get_message(library=lib_id, message="MS_ThisIsaFakeMsg!")

    #Test Assertion Error is handled: Short Library id (get_message: GR)
    def test_responses_long_library_id_get_msg_gr(self):
        '''This method tests that an assertion is raised in the Messages Module's get_message method when the user enters
         a library_id (GR: Global Resource) that is too short. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_gr(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Messages().get_message(library=lib_id, message="MS_ThisIsaFakeMsg!")

    #Test Assertion Error is handled: Incorrect Library id (get_message: GR)
    def test_responses_bad_libary_id_get_msg_gr(self):
        '''This method tests that an assertion is raised in the Messages Module's get_message method when the user enters
         a library_id (GR: Global Resource) that is incorrect. '''
        lib_id, bad_id = setup_tests().setup_test_library_id_gr(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Messages().get_message(library=bad_id, message="MS_ThisIsaFakeMsg!")

    ## Messages: Message IDs ##
    #Test Assertion Error is handled: Long Message ID (get_message)
    def test_responses_short_msg_id_get_msgs(self):
        '''This method tests that an assertion is raised in the Messages Module's get_message method when the user enters a
        message_id that is too long. '''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Messages().get_message(library="GR_ThisIsaFakeLib!", message=msg_id)

    #Test Assertion Error is handled: Short Message ID (get_message)
    def test_responses_long_msg_id_get_msgs(self):
        '''This method tests that an assertion is raised in the Messages Module's get_message method when the user enters a
         message_id that is too short. '''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Messages().get_message(library="GR_ThisIsaFakeLib!", message=msg_id)

    #Test Assertion Error is handled: Incorrect Message ID (get_message)
    def test_responses_bad_msg_id_get_msgs(self):
        '''This method tests that an assertion is raised in the Messages Module's get_message method when the user enters
         a message_id that is incorrect. '''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Messages().get_message(library="GR_ThisIsaFakeLib!", message=msg_id)

    ## Distribution: MailiingList IDs ##
    #Test Assertion Error is handled: Long MailiingListID (create_distribution)
    def test_responses_short_ml_id_create_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters a
        mailing_list_id that is too long. '''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(mailing_list=mailing_list_id, library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", subject="Fake", send_date=gmtime(), reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake", survey="SV_ThisIsaFakeID!!")

    #Test Assertion Error is handled: Short MailiingList ID (create_distribution)
    def test_responses_long_ml_id_create_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters a
         mailing_list_id that is too short.'''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(mailing_list=mailing_list_id, library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", subject="Fake", send_date=gmtime(), reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake", survey="SV_ThisIsaFakeID!!")

    #Test Assertion Error is handled: Incorrect MailiingList ID (create_distribution)
    def test_responses_bad_ml_id_create_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters
         a mailing_list_id that is incorrect. '''
        mailing_list_id, bad_id = setup_tests().setup_test_mailing_list_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(mailing_list=bad_id, library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", subject="Fake", send_date=gmtime(), reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake", survey="SV_ThisIsaFakeID!!")

    ## Distribution: Survey IDs ##
    #Test Assertion Error is handled: Long Survey ID (create_distribution)
    def test_responses_short_sv_id_create_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters a
        survey_id that is too long. '''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(survey=survey_id, mailing_list="CG_ThisIsaFakeMail", library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", send_date=gmtime(), subject="Fake", reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Survey ID (create_distribution)
    def test_responses_long_sv_id_create_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters a
         survey_id that is too short.'''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(survey=survey_id, mailing_list="CG_ThisIsaFakeMail", library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", send_date=gmtime(), subject="Fake", reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Survey ID (create_distribution)
    def test_responses_bad_sv_id_create_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters
         a survey_id that is incorrect. '''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(survey=bad_id, mailing_list="CG_ThisIsaFakeMail", library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", send_date=gmtime(), subject="Fake", reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    ## Distribution: Survey IDs ##
    #Test Assertion Error is handled: Long Survey ID (list_distributions)
    def test_responses_short_sv_id_list_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's list_distributions method when the user enters a
        survey_id that is too long. '''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().list_distributions(survey=survey_id)

    #Test Assertion Error is handled: Short Survey ID (list_distributions)
    def test_responses_long_sv_id_list_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's list_distributions method when the user enters a
         survey_id that is too short.'''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().list_distributions(survey=survey_id)

    #Test Assertion Error is handled: Incorrect Survey ID (list_distributions)
    def test_responses_bad_sv_id_list_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's list_distributions method when the user enters
         a survey_id that is incorrect. '''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().list_distributions(survey=bad_id)

    ## Distribution: Survey IDs ##
    #Test Assertion Error is handled: Long Survey ID (get_distribution)
    def test_responses_short_sv_id_get_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's get_distributionmethod when the user enters a
        survey_id that is too long. '''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().get_distribution(survey=survey_id, distribution="EMD_ThisIsaFakeID!")

    #Test Assertion Error is handled: Short Survey ID (get_distribution)
    def test_responses_long_sv_id_get_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's get_distribution method when the user enters a
         survey_id that is too short.'''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().get_distribution(survey=survey_id, distribution="EMD_ThisIsaFakeID!")

    #Test Assertion Error is handled: Incorrect Survey ID (get_distribution)
    def test_responses_bad_sv_id_get_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's get_distribution method when the user enters
         a survey_id that is incorrect. '''
        survey_id, bad_id = setup_tests().setup_test_survey_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().get_distribution(survey=bad_id, distribution="EMD_ThisIsaFakeID!")

    ## Distribution: Library IDs (UR) ##
    #Test Assertion Error is handled: Long Library ID (create_distribution: UR)
    def test_responses_short_lib_id_create_dist_ur(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters a
        library_id (UR: User-Defined Resource) that is too long. '''
        library_id, bad_id = setup_tests().setup_test_library_id_ur(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(library=library_id, survey="SV_ThisIsaFakeID!!", mailing_list="CG_ThisIsaFakeMail", message="MS_ThisIsaFakeMsg!", send_date=gmtime(), subject="Fake", reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Library ID (create_distribution: UR)
    def test_responses_long_lib_id_create_dist_ur(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters a
         library_id (UR: User-Defined Resource) that is too short.'''
        library_id, bad_id = setup_tests().setup_test_library_id_ur(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(library=library_id, survey="SV_ThisIsaFakeID!!", mailing_list="CG_ThisIsaFakeMail", message="MS_ThisIsaFakeMsg!", send_date=gmtime(), subject="Fake", reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Library ID (create_distribution: UR)
    def test_responses_bad_lib_id_create_dist_ur(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters
         a library_id (UR: User-Defined Resource) that is incorrect. '''
        library_id, bad_id = setup_tests().setup_test_library_id_ur(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(library=bad_id, survey="SV_ThisIsaFakeID!!", mailing_list="CG_ThisIsaFakeMail", message="MS_ThisIsaFakeMsg!", send_date=gmtime(), subject="Fake", reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    ## Distribution: Library IDs (GR) ##
    #Test Assertion Error is handled: Long Library ID (create_distribution: GR)
    def test_responses_short_lib_id_create_dist_gr(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters a
        library_id (GR: Global Resource) that is too long. '''
        library_id, bad_id = setup_tests().setup_test_library_id_gr(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(library=library_id, survey="SV_ThisIsaFakeID!!", mailing_list="CG_ThisIsaFakeMail", message="MS_ThisIsaFakeMsg!", send_date=gmtime(), subject="Fake", reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Library ID (create_distribution: GR)
    def test_responses_long_lib_id_create_dist_gr(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters a
         library_id (GR: Global Resource) that is too short.'''
        library_id, bad_id = setup_tests().setup_test_library_id_gr(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(library=library_id, survey="SV_ThisIsaFakeID!!", mailing_list="CG_ThisIsaFakeMail", message="MS_ThisIsaFakeMsg!", send_date=gmtime(), subject="Fake", reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Library ID (create_distribution: GR)
    def test_responses_bad_lib_id_create_dist_gr(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters
         a library_id (GR: Global Resource) that is incorrect. '''
        library_id, bad_id = setup_tests().setup_test_library_id_gr(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(library=bad_id, survey="SV_ThisIsaFakeID!!", mailing_list="CG_ThisIsaFakeMail", message="MS_ThisIsaFakeMsg!", send_date=gmtime(), subject="Fake", reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    ## Distribution: Library IDs (UR) ##
    #Test Assertion Error is handled: Long Library ID (create_reminder: UR)
    def test_responses_short_lib_id_create_remind_ur(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters a
        library_id (UR: User-Defined Resource) that is too long. '''
        library_id, bad_id = setup_tests().setup_test_library_id_ur(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(library=library_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Library ID (create_reminder: UR)
    def test_responses_long_lib_id_create_remind_ur(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters a
         library_id (UR: User-Defined Resource) that is too short.'''
        library_id, bad_id = setup_tests().setup_test_library_id_ur(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(library=library_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Library ID (create_reminder: UR)
    def test_responses_bad_lib_id_create_remind_ur(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters
         a library_id (UR: User-Defined Resource) that is incorrect. '''
        library_id, bad_id = setup_tests().setup_test_library_id_ur(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(library=bad_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    ## Distribution: Library IDs (GR) ##
    #Test Assertion Error is handled: Long Library ID (create_reminder: GR)
    def test_responses_short_lib_id_create_remind_gr(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters a
        library_id (GR: Global Resource) that is too long. '''
        library_id, bad_id = setup_tests().setup_test_library_id_gr(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(library=library_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Library ID (create_reminder: GR)
    def test_responses_long_lib_id_create_remind_gr(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters a
         library_id (GR: Global Resource) that is too short.'''
        library_id, bad_id = setup_tests().setup_test_library_id_gr(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(library=library_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Library ID (create_reminder: GR)
    def test_responses_bad_lib_id_create_remind_gr(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters
         a library_id (GR: Global Resource) that is incorrect. '''
        library_id, bad_id = setup_tests().setup_test_library_id_gr(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(library=bad_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    ## Distribution: Library IDs ##
    #Test Assertion Error is handled: Long Library ID (create_thank_you: UR)
    def test_responses_short_lib_id_create_ty_ur(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters a
        library_id (UR: User-Defined Resource) that is too long. '''
        library_id, bad_id = setup_tests().setup_test_library_id_ur(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(library=library_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Library ID (create_thank_you: UR)
    def test_responses_long_lib_id_create_ty_ur(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters a
         library_id (UR: User-Defined Resource) that is too short.'''
        library_id, bad_id = setup_tests().setup_test_library_id_ur(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(library=library_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Library ID (create_thank_you: UR)
    def test_responses_bad_lib_id_create_ty_ur(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters
         a library_id (UR: User-Defined Resource) that is incorrect. '''
        library_id, bad_id = setup_tests().setup_test_library_id_ur(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(library=bad_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    ## Distribution: Library IDs (GR) ##
    #Test Assertion Error is handled: Long Library ID (create_reminder: GR)
    def test_responses_short_lib_id_create_ty_gr(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters a
        library_id (GR: Global Resource) that is too long. '''
        library_id, bad_id = setup_tests().setup_test_library_id_gr(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(library=library_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Library ID (create_reminder: GR)
    def test_responses_long_lib_id_create_ty_gr(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters a
         library_id (GR: Global Resource) that is too short.'''
        library_id, bad_id = setup_tests().setup_test_library_id_gr(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(library=library_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Library ID (create_reminder: GR)
    def test_responses_bad_lib_id_create_ty_gr(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters
         a library_id (GR: Global Resource) that is incorrect. '''
        library_id, bad_id = setup_tests().setup_test_library_id_gr(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(library=bad_id, distribution="EMD_ThisIsaFakeID!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    ## Distribution: Distribution IDs ##
    #Test Assertion Error is handled: Long Distribution ID (get_distribution)
    def test_responses_short_dist_id_get_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's get_distribution method when the user enters a
        distribution_id that is too long. '''
        dist_id, bad_id = setup_tests().setup_test_dist_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().get_distribution(survey="SV_ThisIsaFakeID!!", distribution=dist_id)

    #Test Assertion Error is handled: Short Distribution ID (get_distribution)
    def test_responses_long_dist_id_get_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's get_distribution method when the user enters a
         distribution_id that is too short.'''
        dist_id, bad_id = setup_tests().setup_test_dist_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().get_distribution(survey="SV_ThisIsaFakeID!!", distribution=dist_id)

    #Test Assertion Error is handled: Incorrect Distribution ID (get_distribution)
    def test_responses_bad_dist_id_get_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's get_distribution method when the user enters
         a distribution_id that is incorrect. '''
        dist_id, bad_id = setup_tests().setup_test_dist_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().get_distribution(survey="SV_ThisIsaFakeID!!", distribution=bad_id)

    ## Distribution. Distribution IDs ##
    #Test Assertion Error is handled: Long Distribution ID (create_reminder)
    def test_responses_short_dist_id_create_remind(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters a
        distrubution_id that is too long. '''
        distribution_id, bad_id = setup_tests().setup_test_dist_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(distribution=distribution_id, library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Distribution ID (create_reminder)
    def test_responses_long_dist_id_create_remind(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters a
         distrubution_id that is too short.'''
        distribution_id, bad_id = setup_tests().setup_test_dist_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(distribution=distribution_id, library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Distribution ID (create_reminder)
    def test_responses_bad_dist_id_create_remind(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters
         a distrubution_id that is incorrect. '''
        distribution_id, bad_id = setup_tests().setup_test_dist_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(distribution=bad_id, library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    ## Distribution: Distribution IDs ##
    #Test Assertion Error is handled: Long Distribution ID (create_thank_you)
    def test_responses_short_dist_id_create_ty(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters a
        distrubution_id that is too long. '''
        distribution_id, bad_id = setup_tests().setup_test_dist_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(distribution=distribution_id, library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Distribution ID (create_thank_you)
    def test_responses_long_dist_id_create_ty(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters a
         distrubution_id that is too short.'''
        distribution_id, bad_id = setup_tests().setup_test_dist_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(distribution=distribution_id, library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Distribution ID (create_thank_you)
    def test_responses_bad_dist_id_create_ty(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters
         a distrubution_id that is incorrect. '''
        distribution_id, bad_id = setup_tests().setup_test_dist_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(distribution=bad_id, library="GR_ThisIsaFakeLib!", message="MS_ThisIsaFakeMsg!", subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    ## Distribution: Message IDs ##
    #Test Assertion Error is handled: Long Message ID (create_distribution)
    def test_responses_short_msg_id_create_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters a
        message_id that is too long. '''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(message=msg_id, library="GR_ThisIsaFakeLib!", survey="SV_ThisIsaFakeID!!", mailing_list="CG_ThisIsaFakeMail", subject="Fake", send_date=gmtime(), reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Message ID (create_distribution)
    def test_responses_long_msg_id_create_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters a
         message_id that is too short.'''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(message=msg_id ,library="GR_ThisIsaFakeLib!", survey="SV_ThisIsaFakeID!!", mailing_list="CG_ThisIsaFakeMail", subject="Fake", send_date=gmtime(), reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Message ID (create_distribution)
    def test_responses_bad_msg_id_create_dist(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_distribution method when the user enters
         a message_id that is incorrect. '''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_distribution(message=bad_id, library="GR_ThisIsaFakeLib!", survey="SV_ThisIsaFakeID!!", mailing_list="CG_ThisIsaFakeMail", subject="Fake", send_date=gmtime(), reply_email="Fake@Fake.com", from_email="Fake@Fake.com", from_name="Fake")

    ## Distribution: Message IDs ##
    #Test Assertion Error is handled: Long Message ID (create_reminder)
    def test_responses_short_msg_id_create_remind(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters a
        message_id that is too long. '''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(distribution="EMD_ThisIsaFakeID!", library="GR_ThisIsaFakeLib!", message=msg_id, subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Message ID (create_reminder)
    def test_responses_long_msg_id_create_remind(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters a
         message_id that is too short.'''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(distribution="EMD_ThisIsaFakeID!", library="GR_ThisIsaFakeLib!", message=msg_id, subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Message ID (create_reminder)
    def test_responses_bad_msg_id_create_remind(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_reminder method when the user enters
         a message_id that is incorrect. '''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_reminder(distribution="EMD_ThisIsaFakeID!", library="GR_ThisIsaFakeLib!", message=bad_id, subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    ## Distribution Message IDs ##
    #Test Assertion Error is handled: Long Message ID (create_thank_you)
    def test_responses_short_msg_id_create_ty(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters a
        message_id that is too long. '''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=False, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(distribution="EMD_ThisIsaFakeID!", library="GR_ThisIsaFakeLib!", message=msg_id, subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Short Message ID (create_thank_you)
    def test_responses_long_msg_id_create_ty(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters a
         message_id that is too short.'''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=True, false_id=False)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(distribution="EMD_ThisIsaFakeID!", library="GR_ThisIsaFakeLib!", message=msg_id, subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    #Test Assertion Error is handled: Incorrect Message ID (create_thank_you)
    def test_responses_bad_msg_id_create_ty(self):
        '''This method tests that an assertion is raised in the Distribution Module's create_thank_you method when the user enters
         a message_id that is incorrect. '''
        msg_id, bad_id = setup_tests().setup_test_message_id(short=False, false_id=True)
        with self.assertRaises(AssertionError):
            Distributions().create_thank_you(distribution="EMD_ThisIsaFakeID!", library="GR_ThisIsaFakeLib!", message=bad_id, subject="Fake", reply_email="Fake@Fake.com", send_date=gmtime(), from_email="Fake@Fake.com", from_name="Fake")

    # #Test Assertion Error is handled: bool first name (Create Contact in XM)
    # def test_bad_first_name_bool_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a first_name that is a type bool.'''
    #     boolFirstName = bool()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(first_name=boolFirstName)

    # #Test Assertion Error is handled: list first name (Create Contact in XM)
    # def test_bad_first_name_list_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a first_name that is a type list.'''
    #     listFirstName = list()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(first_name=listFirstName)

    # #Test Assertion Error is handled: float first name (Create Contact in XM)
    # def test_bad_first_name_float_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a first_name that is a type float.'''
    #     floatFirstName = float()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(first_name=floatFirstName)

    # #Test Assertion Error is handled: int first name (Create Contact in XM)
    # def test_bad_first_name_int_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a first_name that is a type int.'''
    #     intFirstName = int()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(first_name=intFirstName)

    # #Test Assertion Error is handled: dict first name (Create Contact in XM)
    # def test_bad_first_name_dict_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a first_name that is a type dict.'''
    #     dictFirstName = dict()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(first_name=dictFirstName)

    # #Test Assertion Error is handled: tuple first name (Create Contact in XM)
    # def test_bad_first_name_tuple_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a first_name that is a type tuple.'''
    #     tupleFirstName = tuple()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(first_name=tupleFirstName)

    # #Test Assertion Error is handled: bool last name (Create Contact in XM)
    # def test_bad_last_name_bool_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a last_name that is a type bool.'''
    #     boolLastName = bool()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(last_name=boolLastName)

    # #Test Assertion Error is handled: list last name (Create Contact in XM)
    # def test_bad_last_name_list_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a last_name that is a type list.'''
    #     listLastName = list()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(last_name=listLastName)

    # #Test Assertion Error is handled: float last name (Create Contact in XM)
    # def test_bad_last_name_float_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a last_name that is a type float.'''
    #     floatLastName = float()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(last_name=floatLastName)

    # #Test Assertion Error is handled: int last name (Create Contact in XM)
    # def test_bad_last_name_int_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a last_name that is a type int.'''
    #     intLastName = int()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(last_name=intLastName)

    # #Test Assertion Error is handled: dict last name (Create Contact in XM)
    # def test_bad_last_name_dict_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a last_name that is a type dict.'''
    #     dictLastName = dict()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(last_name=dictLastName)

    # #Test Assertion Error is handled: tuple last name (Create Contact in XM)
    # def test_bad_last_name_tuple_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a last_name that is a type tuple.'''
    #     tupleLastName = tuple()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(last_name=tupleLastName)

    # #Test Assertion Error is handled: bool email (Create Contact in XM)
    # def test_bad_email_bool_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a email that is a type bool.'''
    #     boolEmail = bool()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(email=boolEmail)

    # #Test Assertion Error is handled: list email (Create Contact in XM)
    # def test_bad_email_list_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a email that is a type list.'''
    #     listEmail = list()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(email=listEmail)

    # #Test Assertion Error is handled: float email (Create Contact in XM)
    # def test_bad_email_float_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a email that is a type float.'''
    #     floatEmail = float()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(email=floatEmail)

    # #Test Assertion Error is handled: int email (Create Contact in XM)
    # def test_bad_email_int_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a email that is a type int.'''
    #     intEmail = int()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(email=intEmail)

    # #Test Assertion Error is handled: dict email (Create Contact in XM)
    # def test_bad_email_dict_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a email that is a type dict.'''
    #     dictEmail = dict()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(email=dictEmail)

    # #Test Assertion Error is handled: tuple email (Create Contact in XM)
    # def test_bad_email_tuple_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a email that is a type tuple.'''
    #     tupleEmail = tuple()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(email=tupleEmail)

    # #Test Assertion Error is handled: bool phone (Create Contact in XM)
    # def test_bad_phone_bool_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a phone that is a type bool.'''
    #     boolPhone = bool()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(phone=boolPhone)

    # #Test Assertion Error is handled: list phone (Create Contact in XM)
    # def test_bad_phone_list_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a phone that is a type list.'''
    #     listPhone = list()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(phone=listPhone)

    # #Test Assertion Error is handled: float phone (Create Contact in XM)
    # def test_bad_phone_float_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a phone that is a type float.'''
    #     floatPhone = float()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(phone=floatPhone)

    # #Test Assertion Error is handled: int phone (Create Contact in XM)
    # def test_bad_phone_int_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a phone that is a type int.'''
    #     intPhone = int()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(phone=intPhone)

    # #Test Assertion Error is handled: dict phone (Create Contact in XM)
    # def test_bad_phone_dict_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a phone that is a type dict.'''
    #     dictPhone = dict()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(phone=dictPhone)

    # #Test Assertion Error is handled: tuple phone (Create Contact in XM)
    # def test_bad_phone_tuple_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a phone that is a type tuple.'''
    #     tuplePhone = tuple()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(phone=tuplePhone)

    # #Test Assertion Error is handled: bool language (Create Contact in XM)
    # def test_bad_language_bool_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a language that is a type bool.'''
    #     boolLanguage = bool()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(language=boolLanguage)

    # #Test Assertion Error is handled: list language (Create Contact in XM)
    # def test_bad_language_list_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a language that is a type list.'''
    #     listLanguage = list()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(language=listLanguage)

    # #Test Assertion Error is handled: float language (Create Contact in XM)
    # def test_bad_language_float_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a language that is a type float.'''
    #     floatLanguage = float()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(language=floatLanguage)

    # #Test Assertion Error is handled: int language (Create Contact in XM)
    # def test_bad_language_int_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a language that is a type int.'''
    #     intLanguage = int()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(language=intLanguage)

    # #Test Assertion Error is handled: dict language (Create Contact in XM)
    # def test_bad_language_dict_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a language that is a type dict.'''
    #     dictLanguage = dict()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(language=dictLanguage)

    # #Test Assertion Error is handled: tuple language (Create Contact in XM)
    # def test_bad_language_tuple_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a language that is a type tuple.'''
    #     tupleLanguage = tuple()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(language=tupleLanguage)

    # #Test Assertion Error is handled: bool metadata (Create Contact in XM)
    # def test_bad_metadata_bool_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a metadata that is a type bool.'''
    #     boolMetadata = bool()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(metadata=boolMetadata)

    # #Test Assertion Error is handled: list Metadata (Create Contact in XM)
    # def test_bad_metadata_list_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a metadata that is a type list.'''
    #     listMetadata = list()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(metadata=listMetadata)

    # #Test Assertion Error is handled: float Metadata (Create Contact in XM)
    # def test_bad_metadata_float_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a metadata that is a type float.'''
    #     floatMetadata = float()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(metadata=floatMetadata)

    # #Test Assertion Error is handled: int Metadata (Create Contact in XM)
    # def test_bad_metadata_int_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a metadata that is a type int.'''
    #     intMetadata = int()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(metadata=intMetadata)

    # #Test Assertion Error is handled: str Metadata (Create Contact in XM)
    # def test_bad_metadata_str_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a metadata that is a type dict.'''
    #     strMetadata = str()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(metadata=strMetadata)

    # #Test Assertion Error is handled: tuple Metadata (Create Contact in XM)
    # def test_bad_metadata_tuple_create_contact_xm(self):
    #     '''This method tests that an assertion is raised in the XMDirectory Module's create_contact_in_XM method when the user enters
    #      a metadata that is a type tuple.'''
    #     tupleMetadata = tuple()
    #     with self.assertRaises(AssertionError):
    #         XMDirectory().create_contact_in_XM(metadata=tupleMetadata)

if __name__ == "__main__":
    unittest.main()



'''
These HTTP Errors are defined by Qualtrics. Documentation can be found 
(https://api.qualtrics.com/instructions/docs/Instructions/responses.md)

'''

class Qualtrics400Error(Exception):
    '''This Exception handles errors associated with the HTTP 400 (Bad Request) responses.'''
    def __init__(self, msg):
        super().__init__(msg)

class Qualtrics401Error(Exception):
    '''This Exception handles errors associated with the HTTP 401 (Unauthorized) responses.'''
    def __init__(self, msg):
        super().__init__(msg)

class Qualtrics403Error(Exception):
    '''This Exception handles errors associated with the HTTP 403 (Forbidden) responses.'''
    def __init__(self, msg):
        super().__init__(msg)

class Qualtrics429Error(Exception):
    '''This Exception handles errors associated with the HTTP 429 (Too Many Requests) responses.'''
    def __init__(self, msg):
        super().__init__(msg)

class Qualtrics500Error(Exception):
    '''This Exception handles errors associated with the HTTP 500 (Internal Server Error) responses.'''
    def __init__(self, msg):
        super().__init__(msg)

class Qualtrics503Error(Exception):
    '''This Exception handles errors associated with the HTTP 503 (Temporary Internal Server Error) responses.'''
    def __init__(self, msg):
        super().__init__(msg)

class Qualtrics504Error(Exception):
    '''This Exception handles errors associated with the HTTP 504 (Gateway Timeout) responses.'''
    def __init__(self, msg):
        super().__init__(msg)


import time as t
import requests as r
import pandas as pd
from QualtricsAPI.Setup import Credentials
from QualtricsAPI.JSON import Parser
from QualtricsAPI.Exceptions import Qualtrics500Error

class MailingList(Credentials):
    ''' This class contains methods that give users the ability to work with their users Mailing list's and
    their users Mailing Lists contact data within the XMDirectory.'''

    def __init__(self, token=None, directory_id=None, data_center=None):
        self.token = token
        self.data_center = data_center
        self.directory_id = directory_id
        return

    def create_list(self, name=None):
        '''This method will create a mailing list in the XM Directory for the your specified user's account.

        :param list_name: the name of the list to be created.
        :return: tuple containing the Mailing List's and the Mailing List's new id.
        '''
        assert name != None, 'Hey there! The name parameter cannot be None. You need to pass in a new Mailing List name as a string into the name parameter.'
        assert isinstance(name, str) == True, 'Hey there! The name parameter must be of type string.'

        headers, base_url = self.header_setup(content_type=True, xm=True)
        url = f"{base_url}/mailinglists"
        data = {"name": "{0}".format(name)}
        request = r.post(url, json=data, headers=headers)
        response = request.json()
        try:
            list_id = Parser().json_parser(response=response, keys=['id'], arr=False)[0][0]
            return name, list_id
        except:
            print(f"ServerError: {response['meta']['httpStatus']}\nError Code: {response['meta']['error']['errorCode']}\nError Message: {response['meta']['error']['errorMessage']}")

    def list_lists(self, page_size=100):
        '''This method lists all the mailing lists in the directory for the specified user token. You won't typically need to adjust
        the pre-defined parameters.

        :param page_size: The number of mailing lists to return per call.
        :type page_size: int
        :return: A Pandas DataFrame
        '''
        assert page_size != 0, 'Hey there! You need to have a page size greater than 1'

        mailing_lists = pd.DataFrame()
        headers, base_url = self.header_setup(xm=True)
        url = base_url + f"/mailinglists/?pageSize={page_size}"
        try:
          def get_page(mailing_lists=mailing_lists, url=url):
            ''' This method is a nested method that extracts a single page of mailing lists. '''
            request = r.get(url, headers=headers)
            response = request.json()
            keys = ['mailingListId', 'name', 'ownerId', 'lastModifiedDate', 'creationDate','contactCount', 'nextPage']
            lists = Parser().json_parser(response=response, keys=keys, arr=False)
            single_page = pd.DataFrame(lists).transpose()
            single_page.columns = keys
            single_page['creationDate'] = pd.to_datetime(single_page['creationDate'], unit='ms')
            single_page['lastModifiedDate'] = pd.to_datetime(single_page['lastModifiedDate'], unit='ms')
            mailing_lists = pd.concat([mailing_lists, single_page]).reset_index(drop=True)
            next_page = str(response['result']['nextPage'])
            return mailing_lists, next_page, response
          mailing_lists, next_page, response = get_page(mailing_lists=mailing_lists, url=url)
          while next_page != 'None':
            mailing_lists, next_page, response = get_page(mailing_lists=mailing_lists, url=next_page)
          return mailing_lists
        except:
          print(f"ServerError: {response['meta']['httpStatus']}\nError Code: {response['meta']['error']['errorCode']}\nError Message: {response['meta']['error']['errorMessage']}")

    def get_list(self, mailing_list=None):
        '''This function gets the list specfied by the mailing list param and returns the list members.

        :param mailing_list: Your mailing list id that you are interested in getting information on.
        :type mailing_list: str
        :return: A Pandas DataFrame
        '''
        assert mailing_list != None, 'Hey there! The mailing_list parameter cannot be None. You need to pass in a Mailing List ID as a string into the mailing_list parameter.'
        assert isinstance(mailing_list, str) == True, 'Hey there! The mailing_list parameter must be of type string.'
        assert len(mailing_list) == 18, 'Hey, the parameter for "mailing_list" that was passed is the wrong length. It should have 18 characters.'
        assert mailing_list[:3] == 'CG_', 'Hey there! It looks like your Mailing List ID is incorrect. You can find the Mailing List ID on the Qualtrics site under your account settings. It will begin with "CG_". Please try again.'

        headers, base_url = self.header_setup(xm=True)
        url = f"{base_url}/mailinglists/{mailing_list}"
        request = r.get(url, headers=headers)
        response = request.json()
        try:
            list_info = {
                        "mailingListId": response['result']['mailingListId'],
                        "name": response['result']['name'],
                        "ownerId": response['result']['ownerId'],
                        "lastModifiedDate": response['result']['lastModifiedDate'],
                        "creationDate": response['result']['creationDate'],
                        "contactCount": response['result']['contactCount']
            }
            df = pd.DataFrame.from_dict(list_info, orient='index').transpose()
            df['creationDate'] = pd.to_datetime(df['creationDate'], unit='ms')
            df['lastModifiedDate'] = pd.to_datetime(df['lastModifiedDate'], unit='ms')
            return df
        except:
            print(f"ServerError: {response['meta']['httpStatus']}\nError Code: {response['meta']['error']['errorCode']}\nError Message: {response['meta']['error']['errorMessage']}")

    def rename_list(self, mailing_list=None, name=None):
        '''This method takes an existing mailing list name and updates it to reflect the name defined in the name parameter.

        :param mailing_list: the mailing list that you want to rename.
        :type mailing_list: str
        :param name: The new name for the mailing list.
        :type name: str
        :return: A string indicating the success or failure of the method call.
        '''
        assert mailing_list != None, 'Hey there! The mailing_list parameter cannot be None. You need to pass in a Mailing List ID as a string into the mailing_list parameter.'
        assert isinstance(mailing_list, str) == True, 'Hey there! The mailing_list parameter must be of type string.'
        assert name != None, 'Hey there! The name parameter cannot be None. You need to pass in a new Mailing List name as a string into the name parameter.'
        assert isinstance(name, str) == True, 'Hey there! The name parameter must be of type string.'
        assert len(mailing_list) == 18, 'Hey there! The parameter for "mailing_list" that was passed is the wrong length. It should have 18 characters.'
        assert mailing_list[:3] == 'CG_', 'Hey there! It looks like your Mailing List ID is incorrect. You can find the Mailing List ID on the Qualtrics site under your account settings. Please try again.'

        data = {"name": f"{name}"}
        headers, base_url = self.header_setup(content_type=True, xm=True)
        url = f"{base_url}/mailinglists/{mailing_list}"
        request = r.put(url, json=data, headers=headers)
        response = request.json()
        try:
            if response['meta']['httpStatus'] == '200 - OK':
                print(f'Your mailing list "{mailing_list}" has been renamed to {name} in the XM Directory.')
        except:
            print(f"ServerError: {response['meta']['httpStatus']}\nError Code: {response['meta']['error']['errorCode']}\nError Message: {response['meta']['error']['errorMessage']}")
        return

    def delete_list(self, mailing_list=None):
        '''This method will delete the specified mailing list from the given users XM Directory.

        :param mailing_list: Your mailing list id that you are interested in deleting.
        :type mailing_list: str
        :return: A string indicating the success or failure of the method call.
        '''

        assert mailing_list != None, 'Hey, the mailing_list parameter cannot be None. You need to pass in a Mailing List ID as a string into the mailing_list parameter.'
        assert isinstance(mailing_list, str) == True, 'Hey there, the mailing_list parameter must be of type string.'
        assert len(mailing_list) == 18, 'Hey there! The parameter for "mailing_list" that was passed is the wrong length. It should have 18 characters.'
        assert mailing_list[:3] == 'CG_', 'Hey there! It looks like your Mailing List ID is incorrect. You can find the Mailing List ID on the Qualtrics site under your account settings. Please try again.'

        data = {"name": f"{mailing_list}"}
        headers, base_url = self.header_setup(xm=True)
        url = f"{base_url}/mailinglists/{mailing_list}"
        request = r.delete(url, json=data, headers=headers)
        response = request.json()
        try:
            if response['meta']['httpStatus'] == '200 - OK':
                print(f'Your mailing list "{mailing_list}" has been deleted from the XM Directory.')
        except:
            print(f"ServerError: {response['meta']['httpStatus']}\nError Code: {response['meta']['error']['errorCode']}\nError Message: {response['meta']['error']['errorMessage']}")
        return

    def list_contacts(self, mailing_list=None, page_size=100):
        '''This method creates a pandas DataFrame of all the contacts information within the defined mailing list.

        :param mailing_list: the mailing list id
        :type mailing_list: str
        :param page_size: The number of contacts in the mailing list to return per call. (typically this doesn't not need to be changed.)
        :type page_size: int
        :return: A Pandas DataFrame
        '''
        assert len(mailing_list) == 18, 'Hey, the parameter for "mailing_list" that was passed is the wrong length. It should have 18 characters.'
        assert mailing_list[:3] == 'CG_', 'Hey there! It looks like your Mailing List ID is incorrect. You can find the Mailing List ID on the Qualtrics site under your account settings. Please try again.'
        assert page_size != 0, 'Hey there! You need to have a page size greater than 1'
 
        contact_list = pd.DataFrame()
        headers, base_url = self.header_setup(xm=True)
        url = base_url + f"/mailinglists/{mailing_list}/contacts?pageSize={page_size}"
        try:
          def get_page(mailing_list=mailing_list, contact_list=contact_list, url=url):
            request = r.get(url, headers=headers)
            response = request.json()
            keys = ['contactId','firstName','lastName','email','phone','extRef','language','unsubscribed']
            contact_lists = Parser().json_parser(response=response, keys=keys, arr=False)
            single_page = pd.DataFrame(contact_lists).transpose()
            single_page.columns = keys
            single_page['mailing_list'] = mailing_list
            contact_list = pd.concat([contact_list, single_page]).reset_index(drop=True)
            next_page = str(response['result']['nextPage'])
            return contact_list, next_page, response
          contact_list, next_page, response = get_page(mailing_list=mailing_list, contact_list=contact_list, url=url)
          while next_page != 'None':
            contact_list, next_page, response = get_page(mailing_list=mailing_list, contact_list=contact_list, url=next_page)
          return contact_list
        except:
          print(f"ServerError: {response['meta']['httpStatus']}\nError Code: {response['meta']['error']['errorCode']}\nError Message: {response['meta']['error']['errorMessage']}")

    def create_contact_in_list(self, mailing_list=None, **kwargs):
        '''This method creates contacts in the specified mailing list. It is important to remember here that whenever you create a contact in
        a mailing list, you are also creating that contact in the XMDirectory. Once created 2 seperate IDs are created for the contact. The ContactID
        is the reference for the contact in the XMDirectory, and the Contact Lookup ID is the reference of the contact in the Mailing List.

        :param mailing_list: The mailing list id for the list that you want to add the contact too.
        :type mailing_list: str
        :param first_name: The new contact's first name.
        :type first_name: str
        :param last_name: The new contact's last name.
        :type last_name: str
        :param email: The new contact's email.
        :type email: str
        :param phone: The new contact's phone number.
        :tyoe phone: str
        :param external_ref: The new contact's external reference.
        :type external_ref: str
        :param unsubscribed: This parameter denotes whether the new contact is unsubscribed from surveys (Default: False).
        :type unsbscribed: str
        :param language: The language prefered by the new contact (Default: English)
        :type language: str
        :param metadata: Any relevant contact metadata.
        :type metadata: dict
        :return: the contact id (contact_id) in XMDirectory, and the contact id (contact_list_id) in the mailing list.
        '''
        assert len(mailing_list) == 18, 'Hey there! The parameter for "mailing_list" that was passed is the wrong length. It should have 18 characters.'
        assert mailing_list[:3] == 'CG_', 'Hey there! It looks like your Mailing List ID is incorrect. You can find the Mailing List ID on the Qualtrics site under your account settings. Please try again.'

        dynamic_payload = {}
        for key in list(kwargs.keys()):
            assert key in ['first_name', 'last_name', 'email', 'unsubscribed', 'language', 'external_ref', 'metadata', 'phone'], "Hey there! You can only pass in parameters with names in the list, ['first_name', 'last_name', 'email', 'unsubscribed', 'language', 'external_ref', 'metadata']"
            if key == 'first_name':
                dynamic_payload.update({'firstName': kwargs[str(key)]})
            elif key == 'last_name':
                dynamic_payload.update({'lastName': kwargs[str(key)]})
            elif key == 'email':
                dynamic_payload.update({'email': kwargs[str(key)]})
            elif key == 'phone':
                dynamic_payload.update({'phone': kwargs[str(key)]})
            elif key == 'language':
                dynamic_payload.update({'language': kwargs[str(key)]})
            elif key == 'external_ref':
                dynamic_payload.update({'extRef': kwargs[str(key)]})
            elif key == 'unsubscribed':
                dynamic_payload.update({'unsubscribed': kwargs[str(key)]})
            elif key == 'phone':
                dynamic_payload.update({'phone': kwargs[str(key)]})
            elif key == 'metadata':
                assert isinstance(kwargs['metadata'], dict), 'Hey there, your metadata parameter needs to be of type "dict"!'
                dynamic_payload.update({'embeddedData': kwargs[str(key)]})

        headers, base_url = self.header_setup(content_type=True, xm=True)
        url = base_url + f"/mailinglists/{mailing_list}/contacts"
        request = r.post(url, json=dynamic_payload, headers=headers)
        response = request.json()
        try:
            if response['meta']['httpStatus'] == '500 - Internal Server Error':
                raise Qualtrics500Error('500 - Internal Server Error')
            contact_id = response['result']['id']
        except Qualtrics500Error:
            attempt = 0
            while attempt < 20:
                request = r.post(url, json=dynamic_payload, headers=headers)
                response = request.json()
                if response['meta']['httpStatus'] == '500 - Internal Server Error':
                    attempt+=1
                    t.sleep(0.25)
                    continue
                elif response['meta']['httpStatus'] == '200 - OK':
                    return response['result']['id'], response['result']['contactLookupId']
            return print(f"ServerError: {response['meta']['httpStatus']}\nError Code: {response['meta']['error']['errorCode']}\nError Message: {response['meta']['error']['errorMessage']}")
        except Exception:
            print(f"ServerError: {response['meta']['httpStatus']}\nError Code: {response['meta']['error']['errorCode']}\nError Message: {response['meta']['error']['errorMessage']}")
        else:
            contact_list_id = response['result']['contactLookupId']
            return contact_id, contact_list_id